from enum import Enum
from typing import Any, Callable, TypeAlias

ApplyFunc: TypeAlias = Callable[[Any], Any]
ConditionalCheck: TypeAlias = Callable[[Any], bool]
MappingFunc: TypeAlias = Callable[..., dict[str, Any]]


class DROP(Enum):
    """
    A DeleteRelativeObjectPlaceholder (abbrv. DROP) is a placeholder object that indicates
      the object relative to the current value should be dropped. An "object" in this context
      is a dict or a list.

    Examples:

    {   <-- Grandparent (rel to _value)
        'A': {   <-- Parent (rel to _value)
            'B': {      <-- This Object (rel to _value)
                'C': _value
            }
        }
    }

    {   <-- Grandparent (rel to _value1 and _value2)
        'A': [  <-- Parent (rel to _value1 and _value2)
            {       <-- This Object (rel to _value1)
                'B': _value1
            },
            {       <-- This Object (rel to _value2)
                'B': _value2
            }
        ]
    }
    """

    THIS_OBJECT = -1
    PARENT = -2
    GRANDPARENT = -3
    GREATGRANDPARENT = -4


class KEEP:
    """
    A value wrapped in a KEEP object should be ignored by the Mapper class when removing values.

    Partial keeping is _not_ supported (i.e. a KEEP object within an object to be DROP-ed).
    """

    def __init__(self, v: Any):
        self.value = v



"""
This module contains _shared_ global state which is managed by the `Mapper` and used by `get`

Some notes:
- Python globals are at the module-level, and persist per interpeted session
- Python currently uses the GIL which means each process has different globals
    - This means all expected `Mapper` config needs to run within a given process
- When developing, naming conventions should be followed
"""

from dataclasses import dataclass
from typing import Any, Callable, Generic, TypeVar


@dataclass(frozen=True)
class SharedMapperState:
    _trace_len: int
    strict: bool
    custom_dsl_fn: Callable[[dict[str, Any] | list[Any], Any], Any]


K = TypeVar("K")
V = TypeVar("V")


class ImmutableDict(dict, Generic[K, V]):
    """
    Global `dict` that can receive new items, and inserted items are immutable
    """

    def __setitem__(self, key, value):
        if not isinstance(key, str):
            raise ValueError(f"Key type needs to be `str`, got: {type(value)} {value}")
        if not isinstance(value, SharedMapperState):
            raise ValueError(
                f"Value type needs to be `SharedMapperState`, got: {type(value)} {value}"
            )
        if key in self:
            raise ValueError(
                f"Key {key} already exists and ImmutableDict doesn't allow updating keys."
            )
        super().__setitem__(key, value)

    def update(self, *args, **kwargs):
        raise ValueError("ImmutableDict doesn't allow updates.")

    def __delitem__(self, key):
        raise ValueError("ImmutableDict doesn't allow deletion of keys.")


_Global_Mapper_State_Dict: ImmutableDict[str, SharedMapperState] = ImmutableDict()



from functools import partial
from itertools import islice
from typing import Any, Callable, Container, Iterable, Reversible, TypeVar

import pydian
from pydian.lib.types import DROP, ApplyFunc, ConditionalCheck

"""
`pydian` Wrappers
"""


def get(
    key: str,
    default: Any = None,
    apply: ApplyFunc | Iterable[ApplyFunc] | None = None,
    only_if: ConditionalCheck | None = None,
    drop_level: DROP | None = None,
    flatten: bool | None = None,
):
    """
    Partial wrapper around the Pydian `get` function
    """
    kwargs = {
        "key": key,
        "default": default,
        "apply": apply,
        "only_if": only_if,
        "drop_level": drop_level,
        "flatten": flatten,
    }
    return partial(pydian.get, **kwargs)


"""
Generic Wrappers
"""


def do(func: Callable, *args: Any, **kwargs: Any) -> ApplyFunc:
    """
    Generic partial wrapper for functions.

    Starts at the second parameter when using *args (as opposed to the first).
    """
    return lambda x: func(x, *args, **kwargs)


def add(value: Any, before: bool = False) -> ApplyFunc:
    if before:
        return lambda v: value + v
    return lambda v: v + value


def subtract(value: Any, before: bool = False) -> ApplyFunc:
    if before:
        return lambda v: value - v
    return lambda v: v - value


def multiply(value: Any, before: bool = False) -> ApplyFunc:
    if before:
        return lambda v: value * v
    return lambda v: v * value


def divide(value: Any, before: bool = False) -> ApplyFunc:
    if before:
        return lambda v: value / v
    return lambda v: v / value


T = TypeVar("T", list[Any], tuple[Any])


def keep(n: int) -> ApplyFunc | Callable[[T], T]:
    return lambda it: it[:n]


def index(idx: int) -> ApplyFunc | Callable[[Reversible], Any]:
    def get_index(obj: Reversible, i: int) -> Any:
        if i >= 0:
            it = iter(obj)
        else:
            i = (i + 1) * -1
            it = reversed(obj)
        return next(islice(it, i, i + 1), None)

    return partial(get_index, i=idx)


def equals(value: Any) -> ConditionalCheck:
    return lambda v: v == value


def gt(value: Any) -> ConditionalCheck:
    return lambda v: v > value


def lt(value: Any) -> ConditionalCheck:
    return lambda v: v < value


def gte(value: Any) -> ConditionalCheck:
    return lambda v: v >= value


def lte(value: Any) -> ConditionalCheck:
    return lambda v: v <= value


def equivalent(value: Any) -> ConditionalCheck:
    return lambda v: v is value


def contains(value: Any) -> ConditionalCheck:
    return lambda container: value in container


def contained_in(container: Container) -> ConditionalCheck:
    return lambda v: v in container


def not_equal(value: Any) -> ConditionalCheck:
    return lambda v: v != value


def not_equivalent(value: Any) -> ConditionalCheck:
    return lambda v: v is not value


def not_contains(value: Any) -> ConditionalCheck:
    return lambda container: value not in container


def not_contained_in(container: Container) -> ConditionalCheck:
    return lambda v: v not in container


"""
stdlib Wrappers
"""


def map_to_list(func: Callable) -> ApplyFunc | Callable[[Iterable], list[Any]]:
    """
    Partial wrapper for `map`, then casts to a list
    """
    _map_to_list: Callable = lambda fn, it: list(map(fn, it))
    return partial(_map_to_list, func)


def filter_to_list(func: Callable) -> ApplyFunc | Callable[[Iterable], list[Any]]:
    """
    Partial wrapper for `filter`, then casts to a list
    """
    _filter_to_list: Callable = lambda fn, it: list(filter(fn, it))
    return partial(_filter_to_list, func)
